local Highlighting = FlexBar2.OO.Mixin
{
	"Highlight_Buff",
	"Highlight_Debuff_Gained",
	"Highlight_Debuff_Lost",
	"Highlight_Debuff_Update",
	"Highlight_Gear",
	"Highlight_Spellcast",
	"Highlight_Channel",
	"Highlight_AutoCasting",
	"Highlight_StopAutoCasting",
}

function Highlighting:Activate()
	self:RegisterEvent("SpecialEvents_PlayerBuffGained", "Highlight_Buff");
	self:RegisterEvent("SpecialEvents_PlayerBuffLost", "Highlight_Buff");
	self:RegisterEvent("UNIT_INVENTORY_CHANGED", "Highlight_Gear");
	self:RegisterEvent("SpecialEvents_UnitDebuffGained", "Highlight_Debuff");
	self:RegisterEvent("SpecialEvents_UnitDebuffLost", "Highlight_Debuff");
	self:RegisterEvent("SpecialEvents_AuraTargetChanged", "Highlight_Debuff");
	self:RegisterEvent("SpellCast_Start", "Highlight_Spellcast");
	self:RegisterEvent("SpellCast_Stop", "Highlight_Spellcast");
	self:RegisterEvent("Channeling_Start", "Highlight_Channel");
	self:RegisterEvent("Channeling_Stop", "Highlight_Channel");
	self:RegisterEvent("START_AUTOREPEAT_SPELL","Highlight_AutoCasting");
	self:RegisterEvent("STOP_AUTOREPEAT_SPELL","Highlight_StopAutoCasting");
	self:RegisterEvent("PLAYER_ENTER_COMBAT","Highlight_AutoCasting");
	self:RegisterEvent("PLAYER_LEAVE_COMBAT","Highlight_StopAutoCasting");
	self:Highlight_Gear(nil, nil, "player"); -- Highlight gear as flexbar loads
end

function Highlighting:Highlight_Buff()
	local Texture = self:LoadCurrentTexture(true)
	if(Texture ~= nil and FlexBar2.Aura:UnitHasBuff("player", nil, nil, Texture) ~= nil) then
		self:SetChecked(true, "Buffed");
	else
		self:SetChecked(false, "Buffed");
	end
end

function Highlighting:Highlight_Gear(Event, Frame, Unit)
	if self:GetEffectiveAttribute("type") == "action"
	and HasAction(self:GetEffectiveAttribute("action"))
	and IsEquippedAction(self:GetEffectiveAttribute("action")) == 1 then
		self:SetChecked(true, "Equipped");
	else
		self:SetChecked(false, "Equipped");
	end
end

function Highlighting:Highlight_Debuff()
	if(self:GetEffectiveAttribute("type") == "action" and HasAction(self:GetEffectiveAttribute("action")) and FlexBar2.Aura:UnitHasDebuff("target", nil, nil, GetActionTexture(self:GetEffectiveAttribute("action")))) then
		self:SetChecked(true, "Debuffed");
	else
		self:SetChecked(false, "Debuffed");
	end
end

function Highlighting:Highlight_SpellCast()
	if(self:GetEffectiveAttribute("type") == "action" and HasAction(self:GetEffectiveAttribute("action")) and IsCurrentAction(self:GetEffectiveAttribute("action"))) then
		self:SetChecked(true, "Casting");
	else
		self:SetChecked(false, "Casting");
	end
end

function Highlighting:Highlight_Channel()
	if(self:GetEffectiveAttribute("type") == "action" and HasAction(self:GetEffectiveAttribute("action")) and IsCurrentAction(self:GetEffectiveAttribute("action"))) then
		self:SetChecked(true, "Channeling");
	else
		self:SetChecked(false, "Channeling");
	end
end

function Highlighting:Highlight_AutoCasting()
	local actionID = self:GetEffectiveAttribute("action")
	if HasAction(actionID) and (IsAutoRepeatAction(actionID) or IsAttackAction(actionID)) then
		self:SetChecked(true, "AutoCast");
	else
		self:SetChecked(false, "AutoCast");
	end
end

function Highlighting:Highlight_StopAutoCasting()
	local actionID = self:GetEffectiveAttribute("action")
	if HasAction(actionID) then
		self:SetChecked(false, "AutoCast");
	end
end

FlexBar2:RegisterHandler(Highlighting);
