
function FlexBar2:GetStandardMiscHeaderAttribs()
	return {
	-- set statebutton in the header and have the button pull data from here...
		{"statebutton","0:normal"},
	-- Set the current state
		{"state","0"},
	}
end

function FlexBar2:GetStandardRogueHeaderAttribs()
	return {
		{"statemap-stealth-0","0"}, -- Unstealthed
		{"statemap-stealth-1","1"}, -- Stealthed
	-- set statebutton in the header and have the button pull data from here...
		{"statebutton","0:normal;1:stealth"},
	-- Set the current state
		{"state",tostring(GetShapeshiftForm(true))},
	}
end

function FlexBar2:GetStandardPriestHeaderAttribs()
	return {
		{"statemap-stance-0","0"}, -- Normal
		{"statemap-stance-1","1"}, -- Shadowform
	-- set statebutton in the header and have the button pull data from here...
		{"statebutton","0:normal;1:shadow"},
	-- Set the current state
		{"state",tostring(GetShapeshiftForm(true))},
	}
end

function FlexBar2:GetStandardWarriorHeaderAttribs()
	return {
		{"statemap-stance-1","0"}, -- Battle -- is the Warrior normal stance
		{"statemap-stance-2","2"}, -- Defense -- This may be Berserker stance on a Warrior without Defense stance... but that shouldn't be a problem in practice
		{"statemap-stance-3","3"}, -- Berserker
	-- set statebutton in the header and have the button pull data from here...
		{"statebutton","0:normal;2:defense;3:berserk"},
	-- Set the current state
		{"state",tostring(GetShapeshiftForm(true))},
	}
end

function FlexBar2:GetStandardDruidHeaderAttribs(header)
	-- TODO Actually Verify all of this and ensure that the statebutton contents for Druids below is accurate.
	return {
		{"statemap-stance-0","0"}, -- Normal
		{"statemap-stance-1","4"}, -- Bear/ Dire Bear
		{"statemap-stance-2","5"}, -- Aquatic
		{"statemap-stance-3","6"}, -- Cat
		{"statemap-stance-4","7"}, -- Travel
		{"statemap-stance-5","8"}, -- Moonkin Form
		{"statemap-stance-6","9"}, -- Not 100% sure about these... will need a high level druid to test this stuff...
		{"statemap-stance-7","10"},
	-- the following makes sure that losing stealth only changes the bar if you remained in cat form, otherwise the other stance stuff takes effect.
		{"statemap-stealth-0","1:6"}, -- prowl (1) to cat (6)
		{"statemap-stealth-1","6:1"}, -- cat to prowl
	-- set the header's active state upon creation.
	-- TODO Set the header state for druids appropriately.
	-- set statebutton in the header and have the button pull data from here...
		{"statebutton","0:normal;1:stealth;4:bear;5:aquatic;6:cat;7:travel;8:moonkin"},
	}
end

-- call this like: FlexBar2.Group.Handlers.SetStanceHeaderByClass(GroupName,[true])
FlexBar2.Group:RegisterHandler("SetStanceHeaderByClass",function(self,Save)
	local groupName = self
	if type(self) == "table" then
		if self.Group then
			groupName = self.Group.name
		else
			groupName = self.name
		end
	end
	local class = select(2,UnitClass("player"))
	if class == "DRUID" then
		FlexBar2.Group:DefineSecureHeader(groupName,"SecureStateDriverTemplate",FlexBar2:GetStandardDruidHeaderAttribs())
	elseif class == "ROGUE" then
		FlexBar2.Group:DefineSecureHeader(groupName,"SecureStateDriverTemplate",FlexBar2:GetStandardRogueHeaderAttribs())
	elseif class == "WARRIOR" then
		FlexBar2.Group:DefineSecureHeader(groupName,"SecureStateDriverTemplate",FlexBar2:GetStandardWarriorHeaderAttribs())
	else
		FlexBar2.Group:DefineSecureHeader(groupName,"SecureStateDriverTemplate",FlexBar2:GetStandardMiscHeaderAttribs())
	end
	if Save then
		FlexBar2.db.profile.Groups[groupName].autoHandler = FlexBar2.db.profile.Groups[groupName].autoHandler or {}
		table.insert(FlexBar2.db.profile.Groups[groupName].autoHandler,"SetStanceHeaderByClass")
	end
end)

-- the following are tables which can be used with the SetStanceHeaderByConfig desc line. Additional configurations are also possible.
stealthstancebase = {
	{name="stealth", min=0, max=1, short="S", [0]=0, [1]=1, statefunc="IsStealthed"},
} -- results in states:  S0 and S1

prieststancebase = {
	{name="stance", min=0, max=1, short="F", [0]=0, [1]=1, statefunc="GetShapeShiftForm", statefuncarg=true},
} -- results in states:  F0 and F1

warriorstancebase = {
	{name="stance", min=1, max=3, short="F", [1]=1, [2]=2, [3]=3, statefunc="GetShapeShiftForm", statefuncarg=true},
} -- results in states:  F1 F2 and F3

druidstancebase = { -- Note that druids may benefit from mapping state 2 and 4 to bar 0 as shown next line.  those are travel forms without extra actions.
	{name="stance", min=0, max=7, short="F", [0]=0, [1]=1, [2]=0, [3]=3, [4]=0, [5]=5, [6]=6, [7]=7,
		in_3={{name="stealth", min=0, max=1, short="S", [0]=0, [1]=1, default=0, statefunc="IsStealthed"}},
		statefunc="GetShapeShiftForm", statefuncarg=true
	},
} -- results in states:  F0 F1 F3 F3S1 F5 F6 and F7  (F3S1 is Prowl, the rest map directly to individual forms)

-- generates a list of states from the description, and also generates prefixes and suffixes.
local function GenerateStateList(desc,i,min)
	min = (min or 1) - 1
	i = i or #desc
	if i == min then return {""} end
	local state = desc[i]
	local t = {}
	for s = state.min, state.max do
		for n,ss in ipairs(GenerateStateList(desc,i-1,min+1)) do
			if state["in_"..s] then
				for x,is in ipairs(GenerateStateList(state["in_"..s],nil)) do
					t[#t+1] = ss..state.short..s..is
				end
			else
				if s == state.default then
					t[#t+1] = ss
				else
					t[#t+1] = ss..state.short..s
				end
			end
		end
	end
	return t
end

-- used by GenerateStateMap to generate the state transitions
local function GenerateTransitions(desc,i,n,parent)
	parent = parent or ""
	local prefixstates = GenerateStateList(desc,i-1,1)
	local suffixstates = GenerateStateList(desc,nil,i+1)
	local t = {}
	for a, prefix in ipairs(prefixstates) do
		for b, suffix in ipairs(suffixstates) do
			local from = {}
			local to
			if n == desc[i].default then
				to = parent..prefix..suffix
			else
				to = parent..prefix..desc[i].short..desc[i][n]..suffix
			end
			for c = desc[i].min, desc[i].max do
				if c ~= n and desc[i][c] == c and desc[i][c] ~= desc[i][n] then
					local tfrom
					if c == desc[i].default then
						tfrom = parent..prefix..suffix
					else
						tfrom = parent..prefix..desc[i].short..desc[i][c]..suffix
					end
					from[#from+1] = tfrom
				end
			end
			from = table.concat(from,",")
			t[#t+1] = from..":"..to
		end
	end
	return table.concat(t,";")
end

-- this function is used by GenerateStateAttributes to create the stancemap- entries.
local function GenerateStateMap(desc,t,parent)
	local t = t or {}
	for i,state in ipairs(desc) do
		for n = state.min,state.max do
			t[#t+1] = {"statemap-"..state.name.."-"..n,GenerateTransitions(desc,i,n,parent)}
			if state["in_"..n] then
				GenerateStateMap(state["in_"..n],t,state.short..n)
			end
		end
	end
	return t
end

-- generates current state from the settings in the descriptor.
function FlexBar2:GenerateCurrentState(desc,t)
	t = t or {}
	local s = ""
	for i,state in ipairs(desc) do
		local snum
		if state.statefunc then
			snum = getglobal(state.statefunc)(state.statearg)
		end
		if not snum then
			if state.stateconst then
				snum = state.stateconst
			else
				snum = state.min
			end
		end
		s = s..state.short..state[snum]
	end
	t[#t+1] = {"state",s}
	return t
end

-- the following function returns all of the attributes needed to set up a header properly, including initial state.
function FlexBar2:GenerateStateAttributes(desc)
	t = GenerateStateMap(desc)
	local l = GenerateStateList(desc)
	for i,v in ipairs(l) do l[i] = v..":"..v end
	t[#t+1] = {"statebutton",table.concat(l,";")}
	return FlexBar2:GenerateCurrentState(desc,t)
end

FlexBar2.Group:RegisterHandler("SetStanceHeaderByConfig",function(self,desc,Save)
	local groupName = self
	if type(self) == "table" then
		if self.Group then
			groupName = self.Group.name
		else
			groupName = self.name
		end
	end
	FlexBar2.Group:DefineSecureHeader(groupName,"SecureStateDriverTemplate",FlexBar2:GenerateStateAttributes(desc))
	if Save then
		FlexBar2.db.profile.Groups[groupName].autoHandler = FlexBar2.db.profile.Groups[groupName].autoHandler or {}
		table.insert(FlexBar2.db.profile.Groups[groupName].autoHandler,"SetStanceHeaderByConfig")
		FlexBar2.db.profile.Groups[groupName].headerConfig = desc
	end
end)

local Mapping = FlexBar2.OO.Mixin 
{
	"SetNormalMap",
	"SetStanceMap",
};

function Mapping:SetNormalMap(modifier,msButton,bType,id,Save)
	modifier = modifier or ""
	msButton = msButton or ""
	bType = bType or "action"
	local attr = modifier..bType..msButton
	local value = id
	self:SetAttribute("type",bType,Save)
	self:SetAttribute(attr,value,Save)
end

function Mapping:SetStanceMap(modifier,msButton,stance,bType,id,Save)
	modifier = modifier or ""
	msButton = msButton or ""
	stance = stance or "normal"
	bType = bType or "action"
	local attr = modifier..bType..msButton.."-"..stance
	local value = id
	self:SetAttribute("type-"..stance,bType,Save)
	self:SetAttribute(attr,value,Save)
end

FlexBar2:RegisterHandler(Mapping);
