-- Note: this Mixin is a little bit special, it adds both a function in the core and others in the button mixin

function FlexBar2:RegisterTextSub(Name, Triggers, Func) -- The core part of RegisterTextSub (pretty short)
	-- IMPORTANT NOTE: All textsubs must be registered before flexbar runs its OnEnable code !
	if(type(self.TextSubs) ~= "table") then
		self.TextSubs = {};
	end
	if(type(self.TextSubs.FB2_BUTTON_CHANGED) ~= "table") then
		self.TextSubs.FB2_BUTTON_CHANGED = {};
	end
	self.TextSubs.FB2_BUTTON_CHANGED[Name] = Func;
	if(type(Triggers) == "string") then
		if(type(self.TextSubs[Triggers]) ~= "table") then
			self.TextSubs[Triggers] = {};
		end
		self.TextSubs[Triggers][Name] = Func;
	elseif(type(Triggers) == "table") then
		for _, Trigger in ipairs(Triggers) do
			if(type(self.TextSubs[Trigger]) ~= "table") then
				self.TextSubs[Trigger] = {};
			end
			self.TextSubs[Trigger][Name] = Func;
		end
	end
	if(type(self.TextFuncs) ~= "table") then
		self.TextFuncs = {};
	end
	self.TextFuncs[Name] = Func;
end

local TextSubs = FlexBar2.OO.Mixin 
{
	"LoadSubs", -- Load the text subs
	"RegisterTextSubs", -- The button specific part of RegisterTextSub
	"UpdateTextSubs", -- Update certain subs depending on what event triggered this
	"UpdateTextField", -- Update all text subs of one text field
	"SetTextSub", -- Set the text of a sub
};


function TextSubs:Activate()
	local Text1 = self.Frame:CreateFontString(nil, "ARTWORK");
	Text1:SetFontObject("NumberFontNormal");
	Text1:SetJustifyH("LEFT");
	Text1:SetPoint("TOPLEFT", self.Frame, "TOPLEFT");
	self.Frame.Text1 = Text1;

	local Text2 = self.Frame:CreateFontString(nil, "ARTWORK");
	Text2:SetFontObject("NumberFontNormal");
	Text2:SetJustifyH("RIGHT");
	Text2:SetPoint("TOPRIGHT", self.Frame, "TOPRIGHT");
	self.Frame.Text2 = Text2;

	local Text3 = self.Frame:CreateFontString(nil, "ARTWORK");
	Text3:SetFontObject("NumberFontNormal");
	Text3:SetJustifyH("RIGHT");
	Text3:SetPoint("BOTTOMRIGHT", self.Frame, "BOTTOMRIGHT");
	self.Frame.Text3 = Text3;

	local Text4 = self.Frame:CreateFontString(nil, "ARTWORK");
	Text4:SetFontObject("NumberFontNormal");
	Text4:SetJustifyH("LEFT");
	Text4:SetPoint("BOTTOMLEFT", self.Frame, "BOTTOMLEFT");
	self.Frame.Text4 = Text4;

	local Text5 = self.Frame:CreateFontString(nil, "ARTWORK");
	Text5:SetFontObject("NumberFontNormal");
	Text5:SetJustifyH("CENTER");
	Text5:SetPoint("CENTER", self.Frame, "CENTER");
	self.Frame.Text5 = Text5;

	self:RegisterTextSubs();
	self:LoadSubs();
	self:RegisterEvent("FB2_LoadSettings", "LoadSubs");
end

function TextSubs:LoadSubs()
	if(type(self.Settings.TextSubs) == "table") then
		for n,v in ipairs(self.Settings.TextSubs) do
			self:SetTextSub(n, v);
		end
	end
end

function TextSubs:RegisterTextSubs()
	self:RegisterEvent("FB2_BUTTON_CHANGED", "UpdateTextSubs");
	for Triggers, Name in pairs(FlexBar2.TextSubs) do
		if(type(Triggers) == "string") then
			self:RegisterEvent(Triggers, "UpdateTextSubs");
		elseif(type(Triggers) == "table") then
			for n,v in ipairs(Triggers) do
				self:RegisterEvent(v, "UpdateTextSubs");
			end
		else
			return false;
		end
	end
end

function TextSubs:UpdateTextSubs(Event, Frame, ...)
	if(type(self.TextSubs) ~= "table") then
		return;
	end
	for Name, Func in pairs(FlexBar2.TextSubs[Event]) do
		for Field = 1,5 do
			if(type(self.TextSubs[Field]) == "table") then
				for Key, Prefix in ipairs(self.TextSubs[Field].PrefixList) do
					if(Prefix == Name) then
						self.Frame["Text" .. Field]:SetText(Func(self, self.TextSubs[Field].SuffixList[Key], ...));
					end
				end
			end
		end
	end
end

function TextSubs:UpdateTextField(Field)
	for Key, Prefix in ipairs(self.TextSubs[Field].PrefixList) do
		if(type(FlexBar2.TextFuncs[Prefix]) == "function") then
			self.Frame["Text" .. Field]:SetText(FlexBar2.TextFuncs[Prefix](self, self.TextSubs[Field].SuffixList[Key]));
		end
	end
end

function TextSubs:SetTextSub(Field, Text, Save)
	if(Field > 5 or Field < 1 or type(Text) ~= "string") then
		return;
	end
	if(type(self.TextSubs) ~= "table") then
		self.TextSubs = {};
	end
	if(type(self.TextSubs[Field]) ~= "table") then
		self.TextSubs[Field] = {};
	end
	local PrefixList, SuffixList = {}, {};
	for Match in string.gmatch(Text, "%$%((.-)%)") do
		local _, _, Prefix, Suffix = string.find(Match, "(%a*) ?(.*)");
		table.insert(PrefixList, Prefix);
		table.insert(SuffixList, Suffix);
	end
	self.TextSubs[Field].Text = Text;
	self.TextSubs[Field].PrefixList = PrefixList;
	self.TextSubs[Field].SuffixList = SuffixList;
	if(Save) then
		if(type(self.Settings.TextSubs) ~= "table") then
			self.Settings.TextSubs = {};
		end
		self.Settings.TextSubs[Field] = Text;
	end
	self:UpdateTextField(Field);
end

FlexBar2:RegisterHandler(TextSubs);


-- Default TextSub registering --

FlexBar2:RegisterTextSub("itemcount", "BAG_UPDATE", 
function(Button, Item)
	if(Item ~= "") then
		return GetItemCount(Item);
	else
		if(Button:GetEffectiveAttribute("type") == "action" and IsConsumableAction(Button:GetEffectiveAttribute("action"))) then
			return GetActionCount(Button:GetEffectiveAttribute("action"));
		end
	end
end);

FlexBar2:RegisterTextSub("cd", "FB2_UPDATE", 
function(Button)
	if(Button:GetEffectiveAttribute("type") == "action") then
		local start, duration, enable = GetActionCooldown(Button:GetEffectiveAttribute("action"));
		if(duration > 1.5 and start ~= 0) then -- Ignore global cooldown or no cooldown
			local TimeLeft = (duration-(GetTime()-start));
			if(TimeLeft >= 60) then
				return(math.floor ((TimeLeft / 60) + 0.5) .. "m");
			else
				if(TimeLeft == 0) then
					return "";
				else
					return(math.floor (TimeLeft) .. "s");
				end
			end
		end
	end
	return "";
end);

FlexBar2:RegisterTextSub("name", nil,
function(Button)
	return Button.Button or "";
end);

FlexBar2:RegisterTextSub("key", "FB2_BINDINGS_UPDATE",
function(Button)
	if(type(Button.Bindings) == "table") then
		local Binding = Button.Bindings[1];
		if(Binding == "BUTTON3") then
			Binding = "MM";
		end
		return Binding or "";
	else
		return "";
	end
end);

FlexBar2:RegisterTextSub("actiontext", "FB2_ATTR_CACHE_CLEARED",
function(Button)
	if(Button:GetEffectiveAttribute("type") == "action") then
		return GetActionText(Button:GetEffectiveAttribute("action")) or ""
	end
end);

FlexBar2:RegisterTextSub("actionid", "FB2_ATTR_CACHE_CLEARED",
function(Button)
	if(Button:GetEffectiveAttribute("type") == "action") then
		return tostring(Button:GetEffectiveAttribute("action")) or ""
	end
end);
